# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import annotations

__all__ = []

from dataclasses import dataclass, field
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, NoReturn

from astropy.utils.compat.misc import PYTHON_LT_3_10

if TYPE_CHECKING:
    from astropy.cosmology.core import Cosmology


_dataclass_kwargs = {} if PYTHON_LT_3_10 else {"slots": True}


@dataclass(frozen=True, **_dataclass_kwargs)
class ParametersAttribute:
    """Immutable mapping of the Parameters.

    If accessed from the class, this returns a mapping of the Parameter
    objects themselves.  If accessed from an instance, this returns a
    mapping of the values of the Parameters.

    Parameters
    ----------
    attr_name : str
        The name of the class attribute containing the Parameter objects.
    """

    attr_name: str | None = None
    """The name of the class attribute containing the Parameter objects."""

    _name: str = field(init=False)
    """The name of the descriptor on the containing class."""

    def __set_name__(self, owner: Any, name: str) -> None:
        object.__setattr__(self, "_name", name)

    def __get__(
        self, instance: Cosmology | None, owner: type[Cosmology] | None
    ) -> MappingProxyType[str, Any]:
        # called from the class
        if instance is None:
            return getattr(owner, self.attr_name)
        # called from the instance
        return MappingProxyType(
            {n: getattr(instance, n) for n in getattr(instance, self.attr_name)}
        )

    def __set__(self, instance: Any, value: Any) -> NoReturn:
        msg = f"cannot set {self._name!r} of {instance!r}."
        raise AttributeError(msg)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Testing :mod:`astropy.cosmology.parameter._descriptor`."""

from types import MappingProxyType

import pytest

from astropy.cosmology._utils import all_cls_vars
from astropy.cosmology.parameter import Parameter


class ParametersAttributeTestMixin:
    """Test the descriptor for ``parameters`` on Cosmology classes."""

    @pytest.mark.parametrize("name", ["parameters", "_derived_parameters"])
    def test_parameters_from_class(self, cosmo_cls, name):
        """Test descriptor ``parameters`` accessed from the class."""
        descriptor = all_cls_vars(cosmo_cls)[name]
        # test presence
        assert hasattr(cosmo_cls, name)
        # test Parameter is a MappingProxyType
        parameters = getattr(cosmo_cls, name)
        assert isinstance(parameters, MappingProxyType)
        # Test items
        assert set(parameters.keys()) == set(getattr(cosmo_cls, descriptor.attr_name))
        assert all(isinstance(p, Parameter) for p in parameters.values())

    @pytest.mark.parametrize("name", ["parameters", "_derived_parameters"])
    def test_parameters_from_instance(self, cosmo, name):
        """Test descriptor ``parameters`` accessed from the instance."""
        descriptor = all_cls_vars(cosmo)[name]
        # test presence
        assert hasattr(cosmo, name)
        # test Parameter is a MappingProxyType
        parameters = getattr(cosmo, name)
        assert isinstance(parameters, MappingProxyType)
        # Test keys
        assert set(parameters) == set(getattr(cosmo, descriptor.attr_name))

    @pytest.mark.parametrize("name", ["parameters", "_derived_parameters"])
    def test_parameters_cannot_set_on_instance(self, cosmo, name):
        """Test descriptor ``parameters`` cannot be set on the instance."""
        with pytest.raises(AttributeError, match=f"cannot set {name!r} of"):
            setattr(cosmo, name, {})



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""This module defines custom errors and exceptions used in astropy.coordinates.
"""

__all__ = ["ConvertError", "UnknownSiteException"]


# TODO: consider if this should be used to `units`?
class UnitsError(ValueError):
    """
    Raised if units are missing or invalid.
    """


class ConvertError(Exception):
    """
    Raised if a coordinate system cannot be converted to another.
    """


class UnknownSiteException(KeyError):
    def __init__(self, site, attribute, close_names=None):
        message = (
            f"Site '{site}' not in database. Use {attribute} to see available sites."
            f" If '{site}' exists in the online astropy-data repository, use the"
            " 'refresh_cache=True' option to download the latest version."
        )
        if close_names:
            message += " Did you mean one of: '{}'?'".format("', '".join(close_names))
        self.site = site
        self.attribute = attribute
        self.close_names = close_names
        return super().__init__(message)
